/* Top-level object parsing.
   Copyright (C) 2001 Michal Moskal (malekith@pld.org.pl)

This file is part of GCC.

GCC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GCC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */


#include "ksi.h"

const char *storage_name(storage_t s)
{
	switch (s) {
	case storage_local: return "local";
	case storage_public: return "public";
	case storage_extern: return "extern";
	default: return "BAD_STORAGE";
	}
}

storage_t lookup_storage(const char *s)
{
	storage_t r = storage_error;
	
	switch (s[0]) {
	case 'l':
		r = storage_local;
		break;
	case 'p':
		r = storage_public;
		break;
	case 'e':
		r = storage_extern;
		break;
	}

	if (r == storage_error || strcmp(storage_name(r), s) != 0)
		fatal_error("bad storage specifier '%s'", s);
	
	return r;
}

static void parse_func_args(tree *param_list, tree *param_type_list)
{
	struct list_state ctx;
	keyword_t kw;
	tree type, name, arg;
	int halt = 0;

	init_list(&ctx, kw_parm);

	while (!halt) {
		kw = get_keyword_from_list(&ctx);
	
		switch (kw) {
		case kw_nil:
			arg0_term(kw);
			if ((kw = get_keyword_from_list(&ctx)) != kw__error)
				fatal_error("found (%s ...) aftrer (nil) in "
					    "parm list", keyword_name(kw));
		case kw__error:
			*param_type_list = tree_cons(NULL_TREE, void_type_node, 
						     *param_type_list);
			halt++;
			break;
		case kw_elipsis:
			arg0_term(kw);
			if ((kw = get_keyword_from_list(&ctx)) != kw__error)
				fatal_error("found (%s ...) after (elipsis) "
					    "in parm list", keyword_name(kw));
			halt++;
			break;
		case kw_var:
			parse_var(&type, &name);
			arg = build_decl(PARM_DECL, name, type);
			DECL_ARG_TYPE(arg) = type;
			*param_list = chainon(arg, *param_list);
			*param_type_list = tree_cons(NULL_TREE, type, 
						     *param_type_list);
			break;
		default:
			fatal_error("found (%s ...) where (var ...) or (nil) "
				    "was expected", keyword_name(kw));
			break;	
		}
	}

	*param_list = nreverse(*param_list);
	*param_type_list = nreverse(*param_type_list);
}

extern tree dbx_last_input_file;
extern int dbx_last_line, dbx_last_col;

static void compile_body(tree fndecl, keyword_t kw)
{
	tree param_decl, next_param;
	tree body, dummy;

	/* Set line number information.  */
	DECL_SOURCE_FILE(fndecl) = IDENTIFIER_POINTER(dbx_last_input_file);
	DECL_SOURCE_LINE(fndecl) = dbx_last_line;

	/* Announce we are compiling this function.  */
	announce_function(fndecl);

	/* Set up to compile the function and enter it.  */
	current_function_decl = fndecl;
	DECL_INITIAL(fndecl) = error_mark_node;

	pushlevel(0);
	make_decl_rtl(fndecl, 0);

	/* Push all the PARM_DECL nodes onto the current scope (i.e. the 
	 * scope of the subprogram body) so that they can be recognized 
	 * as local variables in the subprogram.   */

	for (param_decl = nreverse(DECL_ARGUMENTS(fndecl));
	     param_decl; param_decl = next_param) {
		next_param = TREE_CHAIN(param_decl);
		TREE_CHAIN(param_decl) = NULL;
		pushdecl(param_decl);
	}

	/* Store back the PARM_DECL nodes. They appear in the right order. */
	DECL_ARGUMENTS(fndecl) = getdecls();

	init_function_start(fndecl, IDENTIFIER_POINTER(dbx_last_input_file), 0);
	expand_function_start(fndecl, 0);
	expand_start_bindings(0);

	/* The only thing in this function is a single return statement, which
	   we now generate.  Note that EXP will have been allocated in
	   permanent_obstack.  To do this properly, we need to enter the 
	   function context before parsing the expression, but then error 
	   recovery is more complex, so we don't do that here.  */

	/* Hm... I wonder if it really *has* to be (bind ...),
	 * or I do sth wrong. If it really does, we can wrap
	 * function in it automagicly. */

	if (kw != kw_bind)
		fatal_error("function body top-expressoion isn't "
			    "(bind ...), but (%s ...)", 
			    keyword_name(kw));
			    
	body = compile(kw, &dummy);

	fini_labels();
	
	if (flag_yydebug)
		debug_tree(body);

	expand_expr_stmt(body);

	/* Now get back out of the function and compile it.  */
	expand_end_bindings(NULL_TREE, 1, 0);

	poplevel(1, 0, 1);

	BLOCK_SUPERCONTEXT(DECL_INITIAL(fndecl)) = fndecl;

	/* Generate rtl for function exit.  */
	expand_function_end(IDENTIFIER_POINTER(dbx_last_input_file), 
			    dbx_last_line, 0);

	/* Run the optimizers and output assembler code for this function. */
	rest_of_compilation(fndecl);

	/* Garbage collect */
	ggc_collect();

	current_function_decl = NULL_TREE;
}

/* func(local|public|extern, RET_TYPE, NAME, ARGS, BODY) */
void parse_func(void)
{
	struct term_state ctx;
	struct arg a;
	storage_t storage;
	keyword_t kw;
	tree ret_type;
	tree name;
	tree fndecl = NULL, fntype;
	tree param_list = NULL_TREE;
	tree param_type_list = NULL_TREE;
	int had_decl = 0;
	
	init_term(&ctx, kw_func);
	
	get_term_arg(&ctx, arg_id, &a);
	storage = lookup_storage(a.u.val_id);
	
	ret_type = parse_type(get_keyword(&ctx));
	
	name = get_id(&ctx);
	
	fndecl = lookup_id(IDENTIFIER_POINTER(name), FUNCTION_DECL);
	
	parse_func_args(&param_list, &param_type_list);

	/* The function type depends on the return type and type of args.  */
	fntype = build_function_type(ret_type, param_type_list);

	if (fndecl) {
		/* check if it agrees */
		same_types(kw_func, TREE_TYPE(fndecl), fntype);
		had_decl++;
	}

	/* first keyword of the body */
	kw = get_keyword(&ctx);
	
	/* Now make the function decl.  */
	if (fndecl == NULL)
		fndecl = build_decl(FUNCTION_DECL, name, fntype);
	if (had_decl) {
		if (DECL_EXTERNAL(fndecl)) {
			error("function `%s' declared `extern' and "
			      "later defined", IDENTIFIER_POINTER(name));
			DECL_EXTERNAL(fndecl) = 0;
		}
		if (TREE_PUBLIC(fndecl) && 
		    storage == storage_local)
		    	error("function `%s' declared `public' and later "
			      "defined `local'", IDENTIFIER_POINTER(name));
		else if (!TREE_PUBLIC(fndecl) &&
		    storage == storage_public)
		    	error("function `%s' declared `local' and later "
			      "defined `public'", IDENTIFIER_POINTER(name));
		TREE_PUBLIC(fndecl) = storage != storage_local;
		if (!TREE_STATIC(fndecl))
			error("redefinition of function `%s'", 
			      IDENTIFIER_POINTER(name));
	} else {
		DECL_EXTERNAL(fndecl) = storage == storage_extern;
		TREE_PUBLIC(fndecl) = storage != storage_local;
		TREE_STATIC(fndecl) = kw == kw_nil;
		DECL_ARGUMENTS(fndecl) = param_list;
		DECL_RESULT(fndecl) = build_decl(RESULT_DECL, NULL_TREE, 
						 ret_type);
		DECL_CONTEXT(DECL_RESULT(fndecl)) = fndecl;
	}

	rest_of_decl_compilation(fndecl, NULL, 1, 0);

	if (!had_decl)
		pushdecl(fndecl);
	
	if (kw == kw_nil) {
		arg0_term(kw);
	} else if (storage == storage_extern)
		error("extern function supplaied with body");
	else
		compile_body(fndecl, kw);
		
	fini_term(&ctx);
}

void parse_global()
{
	struct term_state ctx;
	struct arg a;
	storage_t storage;
	tree type, name, decl;
	
	init_term(&ctx, kw_global);
	
	get_term_arg(&ctx, arg_id, &a);
	storage = lookup_storage(a.u.val_id);

	expect_keyword(&ctx, kw_var);
	parse_var(&type, &name);
	fini_term(&ctx);

	decl = build_decl(VAR_DECL, name, type);
	
	DECL_EXTERNAL(decl) = storage == storage_extern;
	TREE_PUBLIC(decl) = storage != storage_local;
	TREE_STATIC(decl) = storage != storage_extern;
	DECL_CONTEXT(decl) = NULL;
	
	pushdecl(decl);
	
	rest_of_decl_compilation(decl, NULL, 1, 0);
}

void parse_unit()
{
	struct list_state ctx;
	keyword_t kw;

	init_list(&ctx, kw_unit);
	
	while ((kw = get_keyword_from_list(&ctx)) != kw__error) {
		switch (kw) {
		case kw_func:
			parse_func();
			break;
		case kw_global:
			parse_global();
			break;
		case kw_uniondef:
		case kw_structdef:
			parse_structdef(kw == kw_uniondef);
			break;
		case kw_input_file:
			parse_input_file();
			break;
		default:
			fatal_error("(%s ...) cannot be part of unit", 
				    keyword_name(kw));
			break;
		}
	}
}

/****************************************************************
 * Type stuff.
 */

tree bool_type_node;

/*
 * Return 1 if given keyword introduces type expression.
 */
int type_keyword(keyword_t kw)
{
	switch (kw) {
	case kw_int:
	case kw_uint:
	case kw_float:
	case kw_ptr:
	case kw_void:
	case kw_bool:
	case kw_funcptr:
	case kw_union:
	case kw_enum:
	case kw_struct:
		return 1;
	default:
		return 0;
	}
}

tree parse_type(keyword_t kw)
{
	struct term_state ctx;
	struct arg a;
	int size = 0;
	tree t, id;

	init_term(&ctx, kw);

	switch (kw) {
	case kw_bool:
	case kw_void:
		fini_term(&ctx);	/* 0 args */
		return kw == kw_bool ? bool_type_node : void_type_node;
		
	case kw_int:
	case kw_uint:
	case kw_float:
		get_term_arg(&ctx, arg_int|arg_uint, &a);
		size = a.u.val_int.low;
		fini_term(&ctx);
		if (kw == kw_float) {
			switch (size) {
			case FLOAT_TYPE_SIZE:
				return float_type_node;
			case DOUBLE_TYPE_SIZE:
				return double_type_node;
			default:
				/* non-integer constant */
				if (size == LONG_DOUBLE_TYPE_SIZE)
					return long_double_type_node;
			    	fatal_error("float type cannot have %d bits", 
					    size);
				return NULL;
			}
		} else {
		       if (size != 8 && size != 16 && size != 32 && 
	        	   size != 64 && size != 128)
			    	fatal_error("int type cannot have %d bits", 
					    size);
			return type_for_size(size, kw == kw_uint);
		}
		
	case kw_ptr:
		t = parse_type(get_keyword(&ctx));
		fini_term(&ctx);
		return build_pointer_type(t);
		
	case kw_struct:
	case kw_union:
		id = get_id(&ctx);
		t = lookup_type(id, kw == kw_struct ? RECORD_TYPE : UNION_TYPE);
		if (t == NULL)
			fatal_error("%s %s undefined",
				    kw == kw_struct ? "struct" : "union",
			            IDENTIFIER_POINTER(id));
		fini_term(&ctx);
		return TREE_TYPE(t);
		
	case kw_funcptr:
		sorry("funcptr");
		fatal_error("cannot continue");
		return NULL;
		
	case kw_enum:
		sorry("enum");
		fatal_error("cannot continue");
		return NULL;
	
	default:
		fatal_error("(%s ...) is not valid type expression",
			    keyword_name(kw));
		return NULL;
	}
}

/* var(type(uint, 32), x) */
void parse_var(tree *type, tree *name)
{
	struct term_state ctx;

	init_term(&ctx, kw_var);
	*type = parse_type(get_keyword(&ctx));
	*name = get_id(&ctx);
	fini_term(&ctx);
	
	if (TYPE_SIZE(*type) == 0)
		fatal_error("trying to use incomplete type");
}

void same_types(keyword_t kw, tree t1, tree t2)
{
	if (TREE_CODE(t1) != TREE_CODE(t2))
		fatal_error("(%s ...) type clash (diffrent kinds)",
			    keyword_name(kw));
		
	switch (TREE_CODE(t1)) {
	case INTEGER_TYPE:
		if (!TREE_UNSIGNED(t1) != !TREE_UNSIGNED(t2))
			fatal_error("(%s ...) type clash (signess)",
				    keyword_name(kw));
		/* fall */
	case REAL_TYPE:
		if (TYPE_PRECISION(t1) != TYPE_PRECISION(t2))
			fatal_error("(%s ...) type clash (type size)",
				    keyword_name(kw));
		break;
	case VOID_TYPE:
	case BOOLEAN_TYPE:
		/* always OK */
		break;
	case POINTER_TYPE:
		same_types(kw, TREE_TYPE(t1), TREE_TYPE(t2));
		break;
	case RECORD_TYPE:
	case UNION_TYPE:
	case ENUMERAL_TYPE:
		/* khm.. we could get it just from typedef, and
		 * there is just single copy of any such animal... */
		if (t1 != t2)
			fatal_error("(%s ...) type clash (diffrent structure, "
				    "union or enum)", keyword_name(kw));
		break;
	
	case FUNCTION_TYPE:
		same_types(kw, TREE_TYPE(t1), TREE_TYPE(t2));
		t1 = TYPE_ARG_TYPES(t1);
		t2 = TYPE_ARG_TYPES(t2);
		while (t1 && t2) {
			same_types(kw, TREE_VALUE(t1), TREE_VALUE(t2));
			t1 = TREE_CHAIN(t1);
			t2 = TREE_CHAIN(t2);
		}
		if (t1 || t2)
			fatal_error("(%s ...) type clash (diffrent number "
				    "of arguments to function)", 
				    keyword_name(kw));
		break;
		
	default:
		ice();
	}
}

/* (structdef NAME FIELDS) */
void parse_structdef(int uniondef)
{
	struct term_state ctx;
	struct list_state ltx;
	tree record, record_name;
	tree decl;
	tree last_field = NULL;

	init_term(&ctx, uniondef ? kw_uniondef : kw_structdef);
	record_name = get_id(&ctx);

	decl = lookup_type(record_name, uniondef ? UNION_TYPE : RECORD_TYPE);
	
	if (decl)
		record = TREE_TYPE(decl);
	else
		record = make_node(uniondef ? UNION_TYPE : RECORD_TYPE);

	/* push it here as we can reference it from structure
	 * definition */
	if (decl == NULL) {
		decl = build_decl(TYPE_DECL, record_name, record);
		pushdecl(decl);
	} else if (TYPE_SIZE(decl)) {
		error("redefinition of %s `%s'", 
		      uniondef ? "union" : "struct",
		      IDENTIFIER_POINTER(record_name));
	}
	
	init_list(&ltx, kw_field);
	
	for (;;) {
		keyword_t kw;
		tree type, name, field;
		
		kw = get_keyword_from_list(&ltx);
		if (kw == kw__error)
			break;
		if (kw != kw_var)
			fatal_error("found (%s ...) in [field...] list, "
				    "instead of expected (var ...)", 
				    keyword_name(kw));
		parse_var(&type, &name);
		field = build_decl(FIELD_DECL, name, type);
		if (last_field)
			TREE_CHAIN(last_field) = field;
		else
			TYPE_FIELDS(record) = field;
		last_field = field;
	}

	fini_term(&ctx);

	TYPE_STUB_DECL(record) = decl;
	
	if (TYPE_FIELDS(record) == NULL) {
		/* forward declaration */
		return;
	}
	
	layout_type(record);
	rest_of_type_compilation(record, global_bindings_p());
}
